%%% Filename : f_frame2rule.pl
%%% Date     : 05.02.2014
%%% Contents : main method to create a rule from a frame condition
%%% Author   : Lara Katharina Spendier <lara@logic.at>

%%% startComputation(+Frame, +Web)
%%%		starts the computation of frame condition into rules
%%%		Frame ... variable that contains the frame condition(s)
%%%		Web   ... 0 if started from commandline, 1 if started from web (i.e. computeweb)
%%%		  	  => important for the printing of "<"
startComputation(Frame, Web):-
	isClasses(Frame, C),
	writef('\nThe frame condition is in the class: %q\n', [C]),
	checkClasses(C, OK),
	( OK = 1 ->
	  writef('\n\nEquivalent Labelled Rule(s): \n'),
	  frames2rules(Frame, C, Rules, Web),
	  nl, 
	  texOut(Frame,Rules,C)
	; writef('\nThe algorithm does not work for this class yet.'),
	  texOut(_,_,100)
	).

%%% isCorrectFormula(+FrameCondition, -CorrectFormula)
%%% 		checks whether there are no free variables in the matrix 
%%%  		and whether the variables in the prefix are unique
%%%		FrameCondition ... FrameCondition given by the user (Prefix : Matrix)
%%%		CorrectFormula ... 0 if it is NOT a correct formula, 1 otherwise
isCorrectFormula(X ';' Y, CF) :-
	isCorrectFormula(X, CF1),
	isCorrectFormula(Y, CF2),
	((CF1 = 0; CF2 = 0) ->
	  CF = 0
	; CF = 1
	).
isCorrectFormula(X ':' Y, CF) :-
	getAllVariablesPrefix(X, XVar),
	getAllVariablesMatrix(Y, YVar),
	containsDoubleVars(XVar, CFX),
  	(CFX = 1 ->
	    CF = 0
	; containsNoFreeVars(XVar, YVar, CF) 
	).
  
%%% getAllVariablesPrefix(+Prefix, -[Variables])
%%%		gets all the variables that occur in the prefix(quantifier part)
%%%		Prefix      ... quantifier part of the formula (before the ':')
%%%		[Variables] ... all variables of the prefix
getAllVariablesPrefix('A' X, VQ) :-
	X = (Var 'E' Y),
	getAllVariablesPrefix('E' Y, VQ1),
	append([Var], VQ1, VQ).
getAllVariablesPrefix('A' X, VQ) :-
	X = (Var 'A' Y),
	getAllVariablesPrefix('A' Y, VQ1),
	append([Var], VQ1, VQ).
getAllVariablesPrefix('A' X, [X]).
getAllVariablesPrefix('E' X, VQ) :-
	X = (Var 'E' Y),
  	getAllVariablesPrefix('E' Y, VQ1),
	append([Var], VQ1, VQ).
getAllVariablesPrefix('E' X, VQ) :-
  	X = (Var 'A' Y),
	getAllVariablesPrefix('A' Y, VQ1),
  	append([Var], VQ1, VQ).
getAllVariablesPrefix('E' X, [X]).

%%% getAllVariablesMatrix(+Matrix, -[Variables])
%%%		gets all the variables that occur in the quantifier-free part (matrix)
%%%		Matrix      ... quantifier-free part of the formula (after the ':')
%%%		[Variables] ... all variables of the matrix
getAllVariablesMatrix(-X, VF) :-
	getAllVariablesMatrix(X, VF).
getAllVariablesMatrix(X & Y, VF) :-
	getAllVariablesMatrix(X, VF1),
	getAllVariablesMatrix(Y, VF2),
	append(VF1, VF2, VF).
getAllVariablesMatrix(X v Y, VF) :-
	getAllVariablesMatrix(X, VF1),
	getAllVariablesMatrix(Y, VF2),
  	append(VF1, VF2, VF).
getAllVariablesMatrix(X -> Y, VF) :-
	getAllVariablesMatrix(X, VF1),
  	getAllVariablesMatrix(Y, VF2),
  	append(VF1, VF2, VF).
getAllVariablesMatrix(X < Y, VF) :-
  	append([X], [Y], VF).

%%% containsDoubleVars(+[PrefixVar], -CDV)
%%% 		checks if the variables in the prefix are (not) unique
%%%		[PrefixVar] ... list of the variables in the prefix
%%%		CDV ... 1 if variables occur (at least) twice in [PrefixVar]
containsDoubleVars([H|T], 1) :-
  	member(H, T).
containsDoubleVars([_|T], CDV) :-
  	containsDoubleVars(T, CDV).
containsDoubleVars([], 0).

%%% containsNoFreeVars(+[PrefixVar], +[MatrixVar], -CFV)
%%% 		checks if there are free variables in the matrix
%%%		[PrefixVar] ... list of the variables in the prefix
%%%		[MatrixVar] ... list of the variables in the matrix
%%%		CDV ... 0 if variables occur in [MatrixVar] but not in [PrefixVar]
containsNoFreeVars(QuantVars, [H|_], 0) :-
	\+ member(H, QuantVars).
containsNoFreeVars(QuantVars, [_|T], CFV) :-
	containsNoFreeVars(QuantVars, T, CFV).
containsNoFreeVars(_, [], 1).
  
%% frames2rules(+Frame, +Class, -Structural, +Web)
%%	splits frame conditions that are concatenated with ';'
%% 	Frame 	... Frame Condition provided by the user
%% 	Class	... class level of the frame condition
%%	Structural ... Structural Rules created out of the frame conditions
%%	Web 	... 0 if started from cmdLine, 1 if started from web
frames2rules([], _, [], _).
frames2rules(X ';' Y, C, L, Web) :-
	C = [CH|CT],
	frame2rule(X, CH, L1, Web),
	frames2rules(Y, CT, L2, Web),
	append(L1, L2, L).
%	append(LR1, LR2, LR).
frames2rules(X, C, L, Web) :-
	C = [Class],
	frame2rule(X, Class, L, Web).

%% frame2rule(+Frame, +Class, -Structural, -Web)
%%	Frame to rule
%% 	Frame 	... Frame provided by the user
%%	Structural ... Structural Rules created out of the axioms
%%	Web 	... 0 if started from cmdLine, 1 if started from web
frame2rule(Frame, C, Structural, Web) :-
	member(C, [p(1), s(1), p(2)]),
	isRules([Frame], C, Structural, Web).
%	Structural = [L1],
%	writef('\n Structural: %p', [Structural]).
%	rewriteStructuralRules(L1, StructuralNew).
%	texOut(Axiom, Analytic, C).

frame2rule(_, C, _, _) :-
	\+ member(C, [p(1), s(1), p(2)]),
	writef('\nThe algorithm does not work for this class yet.'),
	texOut(_,_,100).

%% isRules(+Frame, +Class, -Structural, +Web)
%%	creates the structural rules out of the frame conditions provided by the user
%% 	Frame   ... List of frame conditions (provided by the user)
%%%	Class	... List of classes the frame conditions are in
%%	Structural  ... List of Structural rules according to frame conditions
%%	Web 	... 0 if started from cmdLine, 1 if started from web
isRules([], _, [], _).
isRules([H|T], C, Structural, Web) :-
	frame2structural(H, C, Structural1),
	printrule(Structural1, Web),
	isRules(T, C, Structural2, Web),
	append(Structural1, Structural2, Structural).
