%%% Filename : f_structural.pl
%%% Date     : 06.04.2014
%%% Contents : converts a frame condition into a structural rule by applying the algorithm
%%%	       step(1): applying the invertible rules
%%% Author   : Lara Katharina Spendier <lara@logic.at>

%% frame2structural(+Frame, +Class, -Structural)
%%	creates the structural rule from a frame condition
%%	Frame 	... list of frame conditions provided by the user
%%	Class	... Class the frame condition is in
%%	Structural ... structural rule after applying the algorithm
frame2structural(_ ':' Matrix, Class, Structural) :-
	Class = p(1),
	Con = [[[],[Matrix]]],
%	writef('\n Con: %p', [Con]),
	isStructuralRule(Con, ConFinal),
%	nl,nl,writef('CONFINAL: %q', [ConFinal]),
	removeAxioms(ConFinal, ConFinal1),
%	nl,nl,writef('CONFINAL wo Ax: %q', [ConFinal1]),
	isRuleCompletion(ConFinal1, Class, Structural).
%	nl,nl,writef('Structural: %q', [Structural]).

frame2structural(_ ':' Matrix, Class, Structural) :-
	member(Class, [s(1), p(2)]),
	Con = [[[Matrix],[]]],
%	writef('\n Con: %p', [Con]),
	isStructuralRule(Con, ConFinal),
%	nl,nl,writef('CONFINAL: %q', [ConFinal]),
	removeAxioms(ConFinal, ConFinal1),
%	nl,nl,writef('CONFINAL wo Ax: %q', [ConFinal1]),
	isRuleCompletion(ConFinal1, Class, Structural).
%	nl,nl,writef('Structural: %q', [Structural]).


%% isStructuralRule(+Frame, -ConFinal)
%%	starts step (1) for a list of frame conditions
%% 	Frame	... list of frame conditions provided by the user
%%      ConFinal ... [ [Consequent1: [Antecedent],[Succedent] ], ... ] 
isStructuralRule([],[]).
isStructuralRule([H|T], ConFinal) :-
	applyInvertibility([H], ConFinal1),
	isStructuralRule(T, ConFinal2),
	append(ConFinal1, ConFinal2, ConFinal).

%% removeAxioms(+Rule, -RuleAxiomFree)
%%	removes Rules where we have some axiom (i.e., same variable contained in 
%%	the antecedent and succedent)
removeAxioms([], []).
removeAxioms([[A,S]|T], Rules) :-
	isAxiomRule(A, S, Ax),
	removeAxioms(T, Rules1),
	( Ax = 1 ->
	  Append = []
	; Append = [[A,S]]
	),
	append(Append, Rules1, Rules).

%% isAxiomRule(+Antecedent, +Succedent, -isAxiom)
%%	checks whether a variable of the Antecedent is contained
%%	in the succedent (if yes -> isAxiom = 1)
isAxiomRule([], _, 0).
isAxiomRule([H|T], S, Ax) :-
	( member(H, S) ->
	  Ax = 1
	; isAxiomRule(T, S, Ax)
	).

%% applyInvertibility(+Frame, -ConFinal)
%%	starts application of invertibility to the frame conditions
%%	checks if the frame condition is invertible (isInvertible) and applies either
%%	invertible-right-rules or invertible-left-rules to the list
%%	also considers that some invertible rules create 2 different rules
%% 	Frame	... list of Frame provided by the user
%%      ConFinal ... [Antecedent,Succedent]
applyInvertibility([],[]).

applyInvertibility([[A,S]|T], ConFinal) :-
	isInvertible(S),
	isInvertibleRight([A,S], F),
	append(F, T, CF1),
	applyInvertibility(CF1, ConFinal).

applyInvertibility([[A,S]|T], ConFinal) :-
	isInvertible(A),
	isInvertibleLeft([A,S], F),
	append(F, T, CF1),
	applyInvertibility(CF1, ConFinal).

applyInvertibility([[A,S]|T], ConFinal) :-
	\+ isInvertible(A),
	\+ isInvertible(S),
	applyInvertibility(T, CF1),
	append([[A,S]], CF1, ConFinal).

%% isInvertible(+Formula)
%%	checks if the given formula is invertible (-> defined connectives)
%%	returns false if the given formula is not invertible
%%	Formula	... the formula which could be invertible
isInvertible([_ -> _|_]).
isInvertible([_ & _|_]).
isInvertible([_ v _|_]).
isInvertible([- _|_]).

isInvertible([_|T]) :-
	isInvertible(T).

%% isInvertibleRight(+Consequent, -Final),
%% 	applies the right invertible rules (->,r),(v,r),(&,r) and (-,r)
%%	Consequent	... given consequent of the form [Antecedent, Succedent]
%%	Final		... list of consequents of the form [[A,S],[A,S],...]
%%			    see e.g. (&,r) which creates two rules!
%% (-, r): => - A .. A => 
isInvertibleRight([A, S], F) :-
	member(- Ax1, S),
	remove(- Ax1, S, S1),
	append([Ax1], A, A1),
	F = [[A1, S1]].

%% (->, r): => A -> B .. A => B
isInvertibleRight([A, S], F) :-
	member(Ax1 -> Ax2, S),
	remove(Ax1 -> Ax2, S, S1),
	append([Ax2], S1, S2),
	append([Ax1], A, A1),
	F = [[A1, S2]].

%% (v, r):  => A v B  .. => A, B
isInvertibleRight([A, S], F) :-
	member(Ax1 v Ax2, S),
	remove(Ax1 v Ax2, S, S1), 
	append([Ax1, Ax2], S1, S2),
	F = [[A, S2]].

%% (&, r):  => A & B  .. R1: => A, R2: => B  % generates 2 rules!
isInvertibleRight([A,S], F) :-
	member(Ax1 & Ax2, S),
	remove(Ax1 & Ax2, S, S1),
	append([Ax1], S1, S2),
	append([Ax2], S1, S3),
	F = [[A, S2], [A, S3]].


%% isInvertibleLeft(+Consequent, -Final),
%% 	applies the left invertible rules (->,l),(v,l),(&,l) and (-,l)
%%	Consequent	... given consequent of the form [Antecedent, Succedent]
%%	Final		... list of consequents of the form [[A,S],[A,S],...]
%%			    see e.g. (v,l) which creates two rules!
%% (-, l): - A  => .. R1: => A,
isInvertibleLeft([A,S], F) :-
	member(- Ax1, A), 
	remove(- Ax1, A, A1),
	append([Ax1], S, S1),
	F = [[A1, S1]].

%% (->, l): A -> B => .. R1: => A, R2: B =>
isInvertibleLeft([A,S], F) :-
	member(Ax1 -> Ax2, A), 
	remove(Ax1 -> Ax2, A, A1),
	append([Ax1], S, S1),
	append([Ax2], A1, A2),
	F = [[A1, S1], [A2, S]].

%% (v, l):  A v B =>  .. R1: A =>, R2: B =>
isInvertibleLeft([A,S], F) :-
	member(Ax1 v Ax2, A),
	remove(Ax1 v Ax2, A, A1),
	append([Ax1], A1, A2),
	append([Ax2], A1, A3),
	F = [[A2, S], [A3, S]].

%% (&, l):  A & B =>  .. R1: => A, R2: => B
isInvertibleLeft([A, S], F) :-
	member(Ax1 & Ax2, A),
	remove(Ax1 & Ax2, A, A1), 
	append([Ax1, Ax2], A1, A2),
	F = [[A2, S]].
